
// hybrid simulation
// simple example in C

#include <stdio.h>
#include <stdbool.h>

double time = 0;                // simulation time
// simulation parameters
#define ENDTIME  19
#define STEPSIZE 0.01           // limited accuracy of Euler method
#define MINSTEP  1e-6           // used for bisection

#if 0
//////////////////////////////////////////////////////////////////
// model A definition
#define N 2
double y[N] = { 1, 0 };         // initial state

void dynamic(double *y, double *yin) // model behavior: circle test: y'' = -y
{
    yin[0] = y[1];		// y'
    yin[1] = -y[0];		// y'' = -y
}

// state condition and associated state event
bool cond;			// state condition status
bool eval_state_condition()     // evaluation
{
    return (y[0] > 0.0);
}

void state_event() {
    printf("# state event: time=%f\n", time);
    printf("# ========================================\n");
}

//////////////////////////////////////////////////////////////////
#else
//////////////////////////////////////////////////////////////////
// model B definition
#define N 2
double y[N] = { 0, 0 };         // initial state

double x = 1; // input
void dynamic(double *y, double *yin) // model behavior: y'' = x
{
    yin[0] = y[1];		// y'
    yin[1] = x;		        // y'' = x
}

// state condition and associated state event
bool cond;			// state condition ststus
bool eval_state_condition()     // evaluation
{
    return (y[0] > 1.0);
}

void state_event() {
    x*=-1.15;
    printf("# state event: time=%.8f err=%g\n", time, y[0]-1.0);
    printf("# ========================================\n");
}

#endif
//////////////////////////////////////////////////////////////////

// output
void print() {
    printf("%12.8f % 10.8f % 10.8f %d\n", time, y[0], y[1], (int)cond);
}

double ysave[N];                // saved state
double h = STEPSIZE;            // continuous simulation step

void save_state() {
    for(int i=0; i<N; i++)
        ysave[i]=y[i];
}
void restore_state() {
    for(int i=0; i<N; i++)
        y[i]=ysave[i];
}
void init_state_condition()     // initialization
{
    cond = eval_state_condition();
}

void continuous_step(double *y) // we use Euler method
{
    double yin[N];              // integrators inputs
    dynamic(y,yin);             // evaluate inputs: f(y(t),t)
    for (int i = 0; i < 2; i++)
	y[i] += h * yin[i];     // y(t+h) = y(t) + h * f(y(t),t)
    time += h;                  // advance the simulation time
}

int main()                      // simulation experiment
{
    init_state_condition();	// state condition initial value
    while (time < ENDTIME) {
#if 1
        if(time+h>ENDTIME) h=ENDTIME-time;
#endif
        print();
	double last = time;	// start of step
	save_state();	        // we need saved state for restart
	continuous_step(y);
	bool newcond = eval_state_condition();
	if (newcond != cond) {	// change detected
	    if (h <= MINSTEP) {	// can not shorten the step
		cond = newcond;	// new status of condition
                // ### do state event
                state_event();
		h = STEPSIZE;	// reset step size
	    } else {
		// ### restart step
		time = last;	        // restore time
	        restore_state();        // restore state
		h /= 2;		        // shorten next step
		if (h < MINSTEP)
		    h = MINSTEP;        // limit the step size
		printf("# new step size: h=%g\n", h);
	    }
	}
    }
    print();
    printf("# end of simulation at time=%f\n", time);
    return 0;
}
